@@ -616,6 +616,10 @@ coreconfigitem(
     b'bundle2.debug',
     default=False,
 )
+# which kind of delta to put in the bundled changegroup. Possible value
+# - '': use default behavior
+# - p1: force to always use delta against p1
+# - full: force to always use full content
 coreconfigitem(
     b'devel',
     b'bundle.delta',
@@ -688,9 +692,11 @@ coreconfigitem(
     b'servercafile',
     default=b'',
 )
+# This config option is intended for use in tests only. It is a giant
+# footgun to kill security. Don't define it.
 coreconfigitem(
     b'devel',
-    b'serverexactprotocol',
+    b'server-insecure-exact-protocol',
     default=b'',
 )
 coreconfigitem(
@@ -969,7 +975,7 @@ coreconfigitem(
 coreconfigitem(
     b'experimental',
     b'changegroup3',
-    default=False,
+    default=True,
 )
 coreconfigitem(
     b'experimental',
@@ -1240,6 +1246,11 @@ coreconfigitem(
     b'rust.index',
     default=False,
 )
+coreconfigitem(
+    b'experimental',
+    b'server.allow-hidden-access',
+    default=list,
+)
 coreconfigitem(
     b'experimental',
     b'server.filesdata.recommended-batch-size',
@@ -1285,6 +1296,11 @@ coreconfigitem(
     b'sparse-read.min-gap-size',
     default=b'65K',
 )
+coreconfigitem(
+    b'experimental',
+    b'stream-v3',
+    default=False,
+)
 coreconfigitem(
     b'experimental',
     b'treemanifest',
